@@ -4,14 +4,13 @@ use crate::{
     annostorage::ondisk::AnnoStorageImpl,
     dfs::CycleSafeDFS,
     errors::Result,
-    util::disk_collections::{
-        DiskMap, EvictionStrategy, DEFAULT_BLOCK_CACHE_CAPACITY, DEFAULT_MAX_NUMBER_OF_TABLES,
-    },
+    util::disk_collections::{DiskMap, EvictionStrategy, DEFAULT_BLOCK_CACHE_CAPACITY},
 };
 use itertools::Itertools;
 use rustc_hash::FxHashSet;
-use std::collections::BTreeSet;
+use std::collections::{BTreeSet, HashMap};
 use std::ops::Bound;
+use transient_btree_index::BtreeConfig;
 
 pub const SERIALIZATION_ID: &str = "DiskAdjacencyListV1";
 
@@ -25,17 +24,23 @@ pub struct DiskAdjacencyListStorage {
     stats: Option<GraphStatistic>,
 }
 
-fn get_fan_outs(edges: &DiskMap<Edge, bool>) -> Vec<usize> {
-    let mut fan_outs: Vec<usize> = Vec::new();
-    if !edges.is_empty() {
-        for (_, targets) in &edges.iter().group_by(|(e, _)| e.source) {
-            fan_outs.push(targets.count());
+fn get_fan_outs(edges: &DiskMap<Edge, bool>) -> Result<Vec<usize>> {
+    let mut fan_outs: HashMap<NodeID, usize> = HashMap::default();
+    if !edges.is_empty()? {
+        let all_edges = edges.iter()?;
+        for e in all_edges {
+            let (e, _) = e?;
+            fan_outs
+                .entry(e.source)
+                .and_modify(|num_out| *num_out += 1)
+                .or_insert(1);
         }
     }
     // order the fan-outs
+    let mut fan_outs: Vec<usize> = fan_outs.into_iter().map(|(_, s)| s).collect();
     fan_outs.sort_unstable();
 
-    fan_outs
+    Ok(fan_outs)
 }
 
 impl DiskAdjacencyListStorage {
@@ -58,7 +63,10 @@ impl DiskAdjacencyListStorage {
 }
 
 impl EdgeContainer for DiskAdjacencyListStorage {
-    fn get_outgoing_edges<'a>(&'a self, node: NodeID) -> Box<dyn Iterator<Item = NodeID> + 'a> {
+    fn get_outgoing_edges<'a>(
+        &'a self,
+        node: NodeID,
+    ) -> Box<dyn Iterator<Item = Result<NodeID>> + 'a> {
         let lower_bound = Edge {
             source: node,
             target: NodeID::min_value(),
@@ -70,11 +78,11 @@ impl EdgeContainer for DiskAdjacencyListStorage {
         Box::new(
             self.edges
                 .range(lower_bound..upper_bound)
-                .map(|(e, _)| e.target),
+                .map_ok(|(e, _)| e.target),
         )
     }
 
-    fn has_outgoing_edges(&self, node: NodeID) -> bool {
+    fn has_outgoing_edges(&self, node: NodeID) -> Result<bool> {
         let lower_bound = Edge {
             source: node,
             target: NodeID::min_value(),
@@ -83,10 +91,18 @@ impl EdgeContainer for DiskAdjacencyListStorage {
             source: node,
             target: NodeID::max_value(),
         };
-        self.edges.range(lower_bound..upper_bound).next().is_some()
+        if let Some(edge) = self.edges.range(lower_bound..upper_bound).next() {
+            edge?;
+            Ok(true)
+        } else {
+            Ok(false)
+        }
     }
 
-    fn get_ingoing_edges<'a>(&'a self, node: NodeID) -> Box<dyn Iterator<Item = NodeID> + 'a> {
+    fn get_ingoing_edges<'a>(
+        &'a self,
+        node: NodeID,
+    ) -> Box<dyn Iterator<Item = Result<NodeID>> + 'a> {
         let lower_bound = Edge {
             source: node,
             target: NodeID::min_value(),
@@ -98,13 +114,18 @@ impl EdgeContainer for DiskAdjacencyListStorage {
         Box::new(
             self.inverse_edges
                 .range(lower_bound..upper_bound)
-                .map(|(e, _)| e.target),
+                .map_ok(|(e, _)| e.target),
         )
     }
-    fn source_nodes<'a>(&'a self) -> Box<dyn Iterator<Item = NodeID> + 'a> {
-        let it = self.edges.iter().map(|(e, _)| e.source).unique();
-
-        Box::new(it)
+    fn source_nodes<'a>(&'a self) -> Box<dyn Iterator<Item = Result<NodeID>> + 'a> {
+        match self.edges.iter() {
+            // The unique_by will merge all errors into a single error, which should be ok for our use case
+            Ok(edges) => Box::new(edges.map_ok(|(e, _)| e.source).unique_by(|n| match n {
+                Ok(n) => Some(*n),
+                Err(_) => None,
+            })),
+            Err(e) => Box::new(std::iter::once(Err(e))),
+        }
     }
 
     fn get_statistics(&self) -> Option<&GraphStatistic> {
@@ -135,14 +156,18 @@ impl GraphStorage for DiskAdjacencyListStorage {
             edges: DiskMap::new(
                 Some(&location.join("edges.bin")),
                 EvictionStrategy::default(),
-                Some(DEFAULT_MAX_NUMBER_OF_TABLES),
                 DEFAULT_BLOCK_CACHE_CAPACITY,
+                BtreeConfig::default()
+                    .fixed_key_size(std::mem::size_of::<NodeID>() * 2)
+                    .fixed_value_size(2),
             )?,
             inverse_edges: DiskMap::new(
                 Some(&location.join("inverse_edges.bin")),
                 EvictionStrategy::default(),
-                Some(DEFAULT_MAX_NUMBER_OF_TABLES),
                 DEFAULT_BLOCK_CACHE_CAPACITY,
+                BtreeConfig::default()
+                    .fixed_key_size(std::mem::size_of::<NodeID>() * 2)
+                    .fixed_value_size(2),
             )?,
             annos: AnnoStorageImpl::new(Some(
                 location.join(crate::annostorage::ondisk::SUBFOLDER_NAME),
@@ -171,7 +196,7 @@ impl GraphStorage for DiskAdjacencyListStorage {
         node: NodeID,
         min_distance: usize,
         max_distance: Bound<usize>,
-    ) -> Box<dyn Iterator<Item = NodeID> + 'a> {
+    ) -> Box<dyn Iterator<Item = Result<NodeID>> + 'a> {
         let mut visited = FxHashSet::<NodeID>::default();
         let max_distance = match max_distance {
             Bound::Unbounded => usize::max_value(),
@@ -179,8 +204,8 @@ impl GraphStorage for DiskAdjacencyListStorage {
             Bound::Excluded(max_distance) => max_distance + 1,
         };
         let it = CycleSafeDFS::<'a>::new(self, node, min_distance, max_distance)
-            .map(|x| x.node)
-            .filter(move |n| visited.insert(*n));
+            .map_ok(|x| x.node)
+            .filter_ok(move |n| visited.insert(*n));
         Box::new(it)
     }
 
@@ -189,7 +214,7 @@ impl GraphStorage for DiskAdjacencyListStorage {
         node: NodeID,
         min_distance: usize,
         max_distance: Bound<usize>,
-    ) -> Box<dyn Iterator<Item = NodeID> + 'a> {
+    ) -> Box<dyn Iterator<Item = Result<NodeID>> + 'a> {
         let mut visited = FxHashSet::<NodeID>::default();
         let max_distance = match max_distance {
             Bound::Unbounded => usize::max_value(),
@@ -198,17 +223,23 @@ impl GraphStorage for DiskAdjacencyListStorage {
         };
 
         let it = CycleSafeDFS::<'a>::new_inverse(self, node, min_distance, max_distance)
-            .map(|x| x.node)
-            .filter(move |n| visited.insert(*n));
+            .map_ok(|x| x.node)
+            .filter_ok(move |n| visited.insert(*n));
         Box::new(it)
     }
 
-    fn distance(&self, source: NodeID, target: NodeID) -> Option<usize> {
+    fn distance(&self, source: NodeID, target: NodeID) -> Result<Option<usize>> {
         let mut it = CycleSafeDFS::new(self, source, usize::min_value(), usize::max_value())
-            .filter(|x| target == x.node)
-            .map(|x| x.distance);
+            .filter_ok(|x| target == x.node)
+            .map_ok(|x| x.distance);
 
-        it.next()
+        match it.next() {
+            Some(distance) => {
+                let distance = distance?;
+                Ok(Some(distance))
+            }
+            None => Ok(None),
+        }
     }
     fn is_connected(
         &self,
@@ -216,16 +247,25 @@ impl GraphStorage for DiskAdjacencyListStorage {
         target: NodeID,
         min_distance: usize,
         max_distance: std::ops::Bound<usize>,
-    ) -> bool {
+    ) -> Result<bool> {
         let max_distance = match max_distance {
             Bound::Unbounded => usize::max_value(),
             Bound::Included(max_distance) => max_distance,
             Bound::Excluded(max_distance) => max_distance + 1,
         };
         let mut it = CycleSafeDFS::new(self, source, min_distance, max_distance)
-            .filter(|x| target == x.node);
+            .filter_ok(|x| target == x.node);
 
-        it.next().is_some()
+        match it.next() {
+            Some(next) => {
+                if let Err(e) = next {
+                    Err(e)
+                } else {
+                    Ok(true)
+                }
+            }
+            None => Ok(false),
+        }
     }
 
     fn copy(
@@ -236,17 +276,19 @@ impl GraphStorage for DiskAdjacencyListStorage {
         self.clear()?;
 
         for source in orig.source_nodes() {
+            let source = source?;
             for target in orig.get_outgoing_edges(source) {
+                let target = target?;
                 let e = Edge { source, target };
                 self.add_edge(e.clone())?;
-                for a in orig.get_anno_storage().get_annotations_for_item(&e) {
+                for a in orig.get_anno_storage().get_annotations_for_item(&e)? {
                     self.add_edge_annotation(e.clone(), a)?;
                 }
             }
         }
 
         self.stats = orig.get_statistics().cloned();
-        self.annos.calculate_statistics();
+        self.annos.calculate_statistics()?;
         Ok(())
     }
 
@@ -274,7 +316,7 @@ impl WriteableGraphStorage for DiskAdjacencyListStorage {
     }
 
     fn add_edge_annotation(&mut self, edge: Edge, anno: Annotation) -> Result<()> {
-        if self.edges.get(&edge).is_some() {
+        if self.edges.get(&edge)?.is_some() {
             self.annos.insert(edge, anno)?;
         }
         Ok(())
@@ -284,7 +326,7 @@ impl WriteableGraphStorage for DiskAdjacencyListStorage {
         self.edges.remove(edge)?;
         self.inverse_edges.remove(&edge.inverse())?;
 
-        let annos = self.annos.get_annotations_for_item(edge);
+        let annos = self.annos.get_annotations_for_item(edge)?;
         for a in annos {
             self.annos.remove_annotation_for_item(edge, &a.key)?;
         }
@@ -300,6 +342,7 @@ impl WriteableGraphStorage for DiskAdjacencyListStorage {
         let mut to_delete = std::collections::LinkedList::<Edge>::new();
 
         for target in self.get_outgoing_edges(node) {
+            let target = target?;
             to_delete.push_back(Edge {
                 source: node,
                 target,
@@ -307,6 +350,7 @@ impl WriteableGraphStorage for DiskAdjacencyListStorage {
         }
 
         for source in self.get_ingoing_edges(node) {
+            let source = source?;
             to_delete.push_back(Edge {
                 source,
                 target: node,
@@ -320,7 +364,7 @@ impl WriteableGraphStorage for DiskAdjacencyListStorage {
         Ok(())
     }
 
-    fn calculate_statistics(&mut self) {
+    fn calculate_statistics(&mut self) -> Result<()> {
         let mut stats = GraphStatistic {
             max_depth: 1,
             max_fan_out: 0,
@@ -333,7 +377,7 @@ impl WriteableGraphStorage for DiskAdjacencyListStorage {
             dfs_visit_ratio: 0.0,
         };
 
-        self.annos.calculate_statistics();
+        self.annos.calculate_statistics()?;
 
         let mut has_incoming_edge: BTreeSet<NodeID> = BTreeSet::new();
 
@@ -341,7 +385,8 @@ impl WriteableGraphStorage for DiskAdjacencyListStorage {
         let mut roots: BTreeSet<NodeID> = BTreeSet::new();
         {
             let mut all_nodes: BTreeSet<NodeID> = BTreeSet::new();
-            for (e, _) in self.edges.iter() {
+            for edge in self.edges.iter()? {
+                let (e, _) = edge?;
                 roots.insert(e.source);
                 all_nodes.insert(e.source);
                 all_nodes.insert(e.target);
@@ -357,19 +402,22 @@ impl WriteableGraphStorage for DiskAdjacencyListStorage {
             stats.nodes = all_nodes.len();
         }
 
-        if !self.edges.is_empty() {
-            for (e, _) in self.edges.iter() {
+        let edges_empty = self.edges.is_empty()?;
+
+        if !edges_empty {
+            for edge in self.edges.iter()? {
+                let (e, _) = edge?;
                 roots.remove(&e.target);
             }
         }
 
-        let fan_outs = get_fan_outs(&self.edges);
+        let fan_outs = get_fan_outs(&self.edges)?;
         let sum_fan_out: usize = fan_outs.iter().sum();
 
         if let Some(last) = fan_outs.last() {
             stats.max_fan_out = *last;
         }
-        let inverse_fan_outs = get_fan_outs(&self.inverse_edges);
+        let inverse_fan_outs = get_fan_outs(&self.inverse_edges)?;
 
         // get the percentile value(s)
         // set some default values in case there are not enough elements in the component
@@ -394,13 +442,14 @@ impl WriteableGraphStorage for DiskAdjacencyListStorage {
         }
 
         let mut number_of_visits = 0;
-        if roots.is_empty() && !self.edges.is_empty() {
+        if roots.is_empty() && !edges_empty {
             // if we have edges but no roots at all there must be a cycle
             stats.cyclic = true;
         } else {
             for root_node in &roots {
                 let mut dfs = CycleSafeDFS::new(self, *root_node, 0, usize::max_value());
                 for step in &mut dfs {
+                    let step = step?;
                     number_of_visits += 1;
                     stats.max_depth = std::cmp::max(stats.max_depth, step.distance);
                 }
@@ -424,6 +473,8 @@ impl WriteableGraphStorage for DiskAdjacencyListStorage {
         }
 
         self.stats = Some(stats);
+
+        Ok(())
     }
 
     fn clear(&mut self) -> Result<()> {
@@ -440,8 +491,6 @@ mod tests {
 
     use super::*;
 
-    use itertools::Itertools;
-
     #[test]
     fn multiple_paths_find_range() {
         /*
@@ -501,12 +550,12 @@ mod tests {
         })
         .unwrap();
 
-        let mut found: Vec<NodeID> = gs
+        let found: Result<Vec<NodeID>> = gs
             .find_connected(1, 3, std::ops::Bound::Included(3))
             .collect();
-
+        let mut found = found.unwrap();
         assert_eq!(2, found.len());
-        found.sort();
+        found.sort_unstable();
 
         assert_eq!(4, found[0]);
         assert_eq!(5, found[1]);
@@ -569,30 +618,52 @@ mod tests {
         })
         .unwrap();
 
-        assert_eq!(
-            vec![2, 3],
-            gs.get_outgoing_edges(1).sorted().collect::<Vec<NodeID>>()
-        );
-        assert_eq!(
-            vec![4, 5],
-            gs.get_outgoing_edges(3).sorted().collect::<Vec<NodeID>>()
-        );
-        assert_eq!(0, gs.get_outgoing_edges(6).count());
-        assert_eq!(vec![4], gs.get_outgoing_edges(2).collect::<Vec<NodeID>>());
+        let mut out1 = gs
+            .get_outgoing_edges(1)
+            .collect::<Result<Vec<_>>>()
+            .unwrap();
+        out1.sort_unstable();
+        assert_eq!(vec![2, 3], out1);
 
-        let mut reachable: Vec<NodeID> = gs.find_connected(1, 1, Bound::Included(100)).collect();
-        reachable.sort();
+        let mut out3 = gs
+            .get_outgoing_edges(3)
+            .collect::<Result<Vec<_>>>()
+            .unwrap();
+        out3.sort_unstable();
+        assert_eq!(vec![4, 5], out3);
+
+        let out6 = gs
+            .get_outgoing_edges(6)
+            .collect::<Result<Vec<_>>>()
+            .unwrap();
+        assert_eq!(0, out6.len());
+
+        let out2 = gs
+            .get_outgoing_edges(2)
+            .collect::<Result<Vec<_>>>()
+            .unwrap();
+        assert_eq!(vec![4], out2);
+
+        let reachable: Result<Vec<NodeID>> =
+            gs.find_connected(1, 1, Bound::Included(100)).collect();
+        let mut reachable = reachable.unwrap();
+        reachable.sort_unstable();
         assert_eq!(vec![2, 3, 4, 5, 6, 7], reachable);
 
-        let mut reachable: Vec<NodeID> = gs.find_connected(3, 2, Bound::Included(100)).collect();
-        reachable.sort();
+        let reachable: Result<Vec<NodeID>> =
+            gs.find_connected(3, 2, Bound::Included(100)).collect();
+        let mut reachable = reachable.unwrap();
+        reachable.sort_unstable();
         assert_eq!(vec![6, 7], reachable);
 
-        let mut reachable: Vec<NodeID> = gs.find_connected(1, 2, Bound::Included(4)).collect();
-        reachable.sort();
+        let reachable: Result<Vec<NodeID>> = gs.find_connected(1, 2, Bound::Included(4)).collect();
+        let mut reachable = reachable.unwrap();
+        reachable.sort_unstable();
         assert_eq!(vec![4, 5, 6, 7], reachable);
 
-        let reachable: Vec<NodeID> = gs.find_connected(7, 1, Bound::Included(100)).collect();
+        let reachable: Result<Vec<NodeID>> =
+            gs.find_connected(7, 1, Bound::Included(100)).collect();
+        let reachable = reachable.unwrap();
         assert_eq!(true, reachable.is_empty());
     }
 
@@ -630,7 +701,7 @@ mod tests {
         })
         .unwrap();
 
-        gs.calculate_statistics();
+        gs.calculate_statistics().unwrap();
         assert_eq!(true, gs.get_statistics().is_some());
         let stats = gs.get_statistics().unwrap();
         assert_eq!(true, stats.cyclic);
@@ -866,7 +937,7 @@ mod tests {
         })
         .unwrap();
 
-        gs.calculate_statistics();
+        gs.calculate_statistics().unwrap();
         assert_eq!(true, gs.get_statistics().is_some());
         let stats = gs.get_statistics().unwrap();
         assert_eq!(true, stats.cyclic);
